"""
Do the money dance whenever someone in the sim pays you directly
"""

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.message.message import Block, Message
from hippolyzer.lib.base.templates import MoneyTransactionType, ChatType
from hippolyzer.lib.proxy.addon_utils import send_chat, BaseAddon
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


class PaydayAddon(BaseAddon):
    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        if message.name != "MoneyBalanceReply":
            return
        transaction_block = message["TransactionInfo"][0]
        # Check for direct user -> user transfer
        if transaction_block["TransactionType"] != MoneyTransactionType.GIFT:
            return

        # Check transfer was to us, not from us
        if transaction_block["DestID"] != session.agent_id:
            return
        sender = transaction_block["SourceID"]
        if sender == session.agent_id:
            return

        # Check if they're likely to be in the sim
        sender_obj = region.objects.lookup_avatar(sender)
        if not sender_obj:
            return

        amount = transaction_block['Amount']
        send_chat(
            f"Thanks for the L${amount} secondlife:///app/agent/{sender}/completename !",
            chat_type=ChatType.SHOUT,
        )
        # Do the traditional money dance.
        session.main_region.circuit.send(Message(
            "AgentAnimation",
            Block("AgentData", AgentID=session.agent_id, SessionID=session.id),
            Block("AnimationList", AnimID=UUID("928cae18-e31d-76fd-9cc9-2f55160ff818"), StartAnim=True),
        ))


addons = [PaydayAddon()]




"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
import unittest
from binascii import unhexlify
from uuid import UUID

from hippolyzer.lib.base.datatypes import Vector3
from hippolyzer.lib.base.message.data_packer import TemplateDataPacker
from hippolyzer.lib.base.message.msgtypes import MsgType
from hippolyzer.lib.base.settings import Settings
from hippolyzer.lib.base.message.message import Message, Block
from hippolyzer.lib.base.message.udpdeserializer import UDPMessageDeserializer
from hippolyzer.lib.base.message.udpserializer import UDPMessageSerializer


class TestDeserializer(unittest.TestCase):

    def tearDown(self):
        pass

    def setUp(self):
        self.settings = Settings()
        self.settings.ENABLE_DEFERRED_PACKET_PARSING = False
        self.deserializer = UDPMessageDeserializer(settings=self.settings)

    def test_deserialize(self):
        body = b'\xff\xff\xff\xfb' + b'\x03' + \
               b'\x01\x00\x00\x00' + b'\x02\x00\x00\x00' + b'\x03\x00\x00\x00'
        message = b'\x00' + b'\x00\x00\x00\x01' + b'\x00' + body
        packet = self.deserializer.deserialize(message)

        assert packet.name == 'PacketAck', 'Incorrect deserialization'

    def test_chat(self):
        msg = Message('ChatFromViewer',
                      Block('AgentData', AgentID=UUID('550e8400-e29b-41d4-a716-446655440000'),
                            SessionID=UUID('550e8400-e29b-41d4-a716-446655440000')),
                      Block('ChatData', Message='Hi Locklainn Tester', Type=1, Channel=0))
        serializer = UDPMessageSerializer()
        packed_data = serializer.serialize(msg)

        packet = self.deserializer.deserialize(packed_data)
        data = packet.blocks
        self.assertEqual(data['ChatData'][0].vars['Message'], 'Hi Locklainn Tester')

    def test_simple_zero_coding(self):
        val = b"\x01\x00\x01\x01"
        decoded = self.deserializer.zero_code_expand(val)
        self.assertEqual(b"\x01\x00\x01", decoded)

    def test_wrap_zero_coding(self):
        val = b"\x01\x00\x00\x02\x01"
        decoded = self.deserializer.zero_code_expand(val)
        self.assertEqual(b"\x01" + (b"\x00" * 256) + b"\x00\x00\x01", decoded)

    def test_double_wrap_zero_coding(self):
        val = b"\x01\x00\x00\x00\x02\x01"
        decoded = self.deserializer.zero_code_expand(val)
        self.assertEqual(b"\x01" + (b"\x00" * 256 * 2) + b"\x00\x00\x01", decoded)

    def test_trailing_zero(self):
        val = b"\x00"
        decoded = self.deserializer.zero_code_expand(val)
        self.assertEqual(val, decoded)

    def test_vector3_from_bytes(self):
        # test the 72 byte ObjectUpdate.ObjectData.ObjectData case
        hex_string = '00000000000000000000803f6666da41660000432fffff422233e34100000000000000000000000000000000000000' \
                     '000000000000000000000000000e33de3c000000000000000000000000'
        position = TemplateDataPacker.unpack(unhexlify(hex_string)[16:16 + 12], MsgType.MVT_LLVector3)
        self.assertEqual(position, (128.00155639648438, 127.99840545654297, 28.399967193603516))
        self.assertIsInstance(position, Vector3)




"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import unittest
import re

from hippolyzer.lib.base.message.data import msg_tmpl
from hippolyzer.lib.base.message.template import MessageTemplate, MessageTemplateBlock, MessageTemplateVariable
from hippolyzer.lib.base.message.template_dict import TemplateDictionary
from hippolyzer.lib.base.message.template_parser import MessageTemplateParser
from hippolyzer.lib.base.message.msgtypes import MsgFrequency, MsgEncoding, \
    MsgDeprecation, MsgBlockType, MsgType


class TestDictionary(unittest.TestCase):

    def setUp(self):
        self.template_file = msg_tmpl
        parser = MessageTemplateParser(self.template_file)
        self.template_list = parser.message_templates

    def test_create_dictionary(self):
        TemplateDictionary(None)

    def test_get_packet(self):
        msg_dict = TemplateDictionary(self.template_list)
        packet = msg_dict.get_template_by_name('ConfirmEnableSimulator')
        assert packet is not None, "get_packet failed"
        assert packet.frequency == MsgFrequency.MEDIUM, "Incorrect frequency"
        assert packet.num == 8, "Incorrect message number for ConfirmEnableSimulator"

    def test_get_packet_pair(self):
        msg_dict = TemplateDictionary(self.template_list)
        packet = msg_dict.get_template_by_pair('Medium', 8)
        assert packet.name == 'ConfirmEnableSimulator', "Frequency-Number pair resulting in incorrect packet"


class TestTemplates(unittest.TestCase):
    def setUp(self):
        self.template_file = msg_tmpl
        self.parser = MessageTemplateParser(self.template_file)
        self.msg_dict = TemplateDictionary(self.parser.message_templates)

    def test_parser(self):
        parser = MessageTemplateParser(self.template_file)
        assert parser.message_templates is not None, "Parsing template file failed"

    def test_parser_fail(self):
        with self.assertRaises(Exception):
            _parser = MessageTemplateParser(None)

    def test_parser_version(self):
        version = self.parser.version
        assert version == 2.0, "Version not correct, expected 2.0, got " + str(version)

    def test_template(self):
        template = self.msg_dict['CompletePingCheck']
        name = template.name
        freq = template.frequency
        num = template.num
        trust = template.trusted
        enc = template.encoding
        assert name == 'CompletePingCheck', "Expected: CompletePingCheck  Returned: " + name
        assert freq == MsgFrequency.HIGH, "Expected: High  Returned: " + freq
        assert num == 2, "Expected: 2  Returned: " + str(num)
        assert not trust, "Expected: NotTrusted  Returned: " + trust
        assert enc == MsgEncoding.UNENCODED, "Expected: Unencoded  Returned: " + enc

    def test_deprecated(self):
        template = self.msg_dict['ObjectPosition']
        dep = template.deprecation
        assert dep == MsgDeprecation.DEPRECATED, "Expected:  Deprecated  Returned: " + str(dep)

    def test_template_fixed(self):
        template = self.msg_dict['PacketAck']
        num = template.num
        assert num == 251, "Expected: 251  Returned: " + str(num)

    def test_blacklisted(self):
        template = self.msg_dict['TeleportFinish']
        self.assertEqual(template.deprecation,
                         MsgDeprecation.UDPBLACKLISTED)

    def test_block(self):
        block = self.msg_dict['OpenCircuit'].get_block('CircuitInfo')
        tp = block.block_type
        num = block.number
        assert tp == MsgBlockType.MBT_SINGLE, "Expected:   Single   Returned: " + tp
        assert num == 0, "Expected:   0  Returned: " + str(num)

    def test_block_multiple(self):
        block = self.msg_dict['NeighborList'].get_block('NeighborBlock')
        tp = block.block_type
        num = block.number
        assert tp == MsgBlockType.MBT_MULTIPLE, "Expected:   Multiple   Returned: " + tp
        assert num == 4, "Expected:   4  Returned: " + str(num)

    def test_variable(self):
        variable = self.msg_dict['StartPingCheck'].get_block('PingID').get_variable('PingID')
        tp = variable.type
        assert tp == MsgType.MVT_U8, "Expected: U8  Returned:  " + str(tp)

    def test_add_get_block(self):
        template = MessageTemplate('CompletePingCheck')
        block = MessageTemplateBlock('CircuitCode')
        template.add_block(block)
        blocks = template.blocks
        assert blocks[0].name == 'CircuitCode', "Add block failed"
        assert template.get_block('CircuitCode') is not None, "Get block failed"

    def test_add_variable(self):
        block = MessageTemplateBlock('CircuitCode')
        variable = MessageTemplateVariable("PingID", MsgType.MVT_U8, 1)
        block.add_variable(variable)
        var_list = block.variables
        assert var_list[0].name == 'PingID', "Add variable failed"
        assert block.get_variable('PingID') is not None, "Get variable failed"

    def test_counts(self):
        low_re = r".*?Low\s*?(\d+).*"
        med_re = r".*?Medium\s*?(\d+).*"
        high_re = r".*?High\s*?(\d+).*"
        fixed_re = r".*?Fixed\s*?(0(X|x)\w+).*"
        self.template_file.seek(0)
        lines = self.template_file
        low_count = 0
        medium_count = 0
        high_count = 0
        fixed_count = 0
        while True:
            try:
                line = next(lines)
            except StopIteration:
                break

            low = re.match(low_re, line)
            medium = re.match(med_re, line)
            high = re.match(high_re, line)
            fixed = re.match(fixed_re, line)
            if low:
                low_count += 1
            if medium:
                medium_count += 1
            if high:
                high_count += 1
            if fixed:
                fixed_count += 1

        frequency_counter = {"low": 0, 'medium': 0, "high": 0, 'fixed': 0}
        for template in list(self.msg_dict.message_templates.values()):
            frequency_counter[template.frequency.name.lower()] += 1
        self.assertEqual(low_count, frequency_counter["low"])
        self.assertEqual(medium_count, frequency_counter["medium"])
        self.assertEqual(high_count, frequency_counter["high"])
        self.assertEqual(fixed_count, frequency_counter["fixed"])



"""
Copyright 2007 Pallets
Copyright 2021 Salad Dais

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1.  Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

2.  Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

3.  Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from
    this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import pickle
from copy import copy
from copy import deepcopy

import pytest

from hippolyzer.lib.base.multidict import MultiDict, OrderedMultiDict


class TestNativeItermethods:
    def test_basic(self):
        class StupidDict:
            def keys(self, multi=1):
                return iter(["a", "b", "c"] * multi)

            def values(self, multi=1):
                return iter([1, 2, 3] * multi)

            def items(self, multi=1):
                return iter(
                    zip(iter(self.keys(multi=multi)), iter(self.values(multi=multi)))
                )

        d = StupidDict()
        expected_keys = ["a", "b", "c"]
        expected_values = [1, 2, 3]
        expected_items = list(zip(expected_keys, expected_values))

        assert list(d.keys()) == expected_keys
        assert list(d.values()) == expected_values
        assert list(d.items()) == expected_items

        assert list(d.keys(2)) == expected_keys * 2
        assert list(d.values(2)) == expected_values * 2
        assert list(d.items(2)) == expected_items * 2


class _MutableMultiDictTests:
    storage_class = None

    def test_pickle(self):
        cls = self.storage_class

        def create_instance(module=None):
            if module is None:
                d_inst = cls()
            else:
                old = cls.__module__
                cls.__module__ = module
                d_inst = cls()
                cls.__module__ = old
            d_inst.setlist(b"foo", [1, 2, 3, 4])
            d_inst.setlist(b"bar", b"foo bar baz".split())
            return d_inst

        for protocol in range(pickle.HIGHEST_PROTOCOL + 1):
            d = create_instance()
            s = pickle.dumps(d, protocol)
            ud = pickle.loads(s)
            assert type(ud) is type(d)
            assert ud == d
            alternative = pickle.dumps(create_instance("werkzeug"), protocol)
            assert pickle.loads(alternative) == d
            ud[b"newkey"] = b"bla"
            assert ud != d

    def test_basic_interface(self):
        md = self.storage_class()
        assert isinstance(md, dict)

        mapping = [
            ("a", 1),
            ("b", 2),
            ("a", 2),
            ("d", 3),
            ("a", 1),
            ("a", 3),
            ("d", 4),
            ("c", 3),
        ]
        md = self.storage_class(mapping)

        # simple getitem gives the first value
        assert md["a"] == 1
        assert md["c"] == 3
        with pytest.raises(KeyError):
            md["e"]  # noqa
        assert md.get("a") == 1

        # list getitem
        assert md.getlist("a") == [1, 2, 1, 3]
        assert md.getlist("d") == [3, 4]
        # do not raise if key not found
        assert md.getlist("x") == []

        # simple setitem overwrites all values
        md["a"] = 42
        assert md.getlist("a") == [42]

        # list setitem
        md.setlist("a", [1, 2, 3])
        assert md["a"] == 1
        assert md.getlist("a") == [1, 2, 3]

        # verify that it does not change original lists
        l1 = [1, 2, 3]
        md.setlist("a", l1)
        del l1[:]
        assert md["a"] == 1

        # setdefault, setlistdefault
        assert md.setdefault("u", 23) == 23
        assert md.getlist("u") == [23]
        del md["u"]

        md.setlist("u", [-1, -2])

        # delitem
        del md["u"]
        with pytest.raises(KeyError):
            md["u"]  # noqa
        del md["d"]
        assert md.getlist("d") == []

        # keys, values, items, lists
        assert list(sorted(md.keys())) == ["a", "b", "c"]
        assert list(sorted(md.keys())) == ["a", "b", "c"]

        # Changed from werkzeug. IMO this is less wrong than only
        # returning the first item for each.
        assert list(sorted(md.values())) == [1, 2, 2, 3, 3]
        assert list(sorted(md.values())) == [1, 2, 2, 3, 3]

        assert list(sorted(md.items(multi=False))) == [("a", 1), ("b", 2), ("c", 3)]
        assert list(sorted(md.items())) == [
            ("a", 1),
            ("a", 2),
            ("a", 3),
            ("b", 2),
            ("c", 3),
        ]
        assert list(sorted(md.items(multi=False))) == [("a", 1), ("b", 2), ("c", 3)]
        assert list(sorted(md.items())) == [
            ("a", 1),
            ("a", 2),
            ("a", 3),
            ("b", 2),
            ("c", 3),
        ]

        assert list(sorted(md.lists())) == [("a", [1, 2, 3]), ("b", [2]), ("c", [3])]
        assert list(sorted(md.lists())) == [("a", [1, 2, 3]), ("b", [2]), ("c", [3])]

        # copy method
        c = md.copy()
        assert c["a"] == 1
        assert c.getlist("a") == [1, 2, 3]

        # copy method 2
        c = copy(md)
        assert c["a"] == 1
        assert c.getlist("a") == [1, 2, 3]

        # deepcopy method
        c = md.deepcopy()
        assert c["a"] == 1
        assert c.getlist("a") == [1, 2, 3]

        # deepcopy method 2
        c = deepcopy(md)
        assert c["a"] == 1
        assert c.getlist("a") == [1, 2, 3]

        # update with a multidict
        od = self.storage_class([("a", 4), ("a", 5), ("y", 0)])
        md.update(od)
        assert md.getlist("a") == [1, 2, 3, 4, 5]
        assert md.getlist("y") == [0]

        # update with a regular dict
        md = c
        od = {"a": 4, "y": 0}
        md.update(od)
        assert md.getlist("a") == [1, 2, 3, 4]
        assert md.getlist("y") == [0]

        # pop, poplist, popitem, popitemlist
        assert md.pop("y") == 0
        assert "y" not in md
        assert md.poplist("a") == [1, 2, 3, 4]
        assert "a" not in md
        assert md.poplist("missing") == []

        # remaining: b=2, c=3
        popped = md.popitem()
        assert popped in [("b", 2), ("c", 3)]
        popped = md.popitemlist()
        assert popped in [("b", [2]), ("c", [3])]

        # repr
        md = self.storage_class([("a", 1), ("a", 2), ("b", 3)])
        assert "('a', 1)" in repr(md)
        assert "('a', 2)" in repr(md)
        assert "('b', 3)" in repr(md)

        # add and getlist
        md.add("c", "42")
        md.add("c", "23")
        assert md.getlist("c") == ["42", "23"]
        md.add("c", "blah")
        assert md.getlist("c", elem_type=int) == [42, 23]

        # setdefault
        md = self.storage_class()
        md.setdefault("x", []).append(42)
        md.setdefault("x", []).append(23)
        assert md["x"] == [42, 23]

        # to dict
        md = self.storage_class()
        md["foo"] = 42
        md.add("bar", 1)
        md.add("bar", 2)
        assert md.to_dict() == {"foo": 42, "bar": 1}
        assert md.to_dict(flat=False) == {"foo": [42], "bar": [1, 2]}

        # popitem from empty dict
        with pytest.raises(KeyError):
            self.storage_class().popitem()

        with pytest.raises(KeyError):
            self.storage_class().popitemlist()

        # key errors are of a special type
        with pytest.raises(KeyError):
            self.storage_class()[42]  # noqa

        # setlist works
        md = self.storage_class()
        md["foo"] = 42
        md.setlist("foo", [1, 2])
        assert md.getlist("foo") == [1, 2]


class TestMultiDict(_MutableMultiDictTests):
    storage_class = MultiDict  # type: ignore

    def test_multidict_pop(self):
        def make_d():
            return self.storage_class({"foo": [1, 2, 3, 4]})

        d = make_d()
        assert d.pop("foo") == 1
        assert not d
        d = make_d()
        assert d.pop("foo", 32) == 1
        assert not d
        d = make_d()
        assert d.pop("foos", 32) == 32
        assert d

        with pytest.raises(KeyError):
            d.pop("foos")

    def test_multidict_pop_raise_keyerror_for_empty_list_value(self):
        mapping = [("a", "b"), ("a", "c")]
        md = self.storage_class(mapping)

        md.setlistdefault("empty", [])

        with pytest.raises(KeyError):
            md.pop("empty")

    def test_multidict_popitem_raise_keyerror_for_empty_list_value(self):
        mapping = []
        md = self.storage_class(mapping)

        md.setlistdefault("empty", [])

        with pytest.raises(KeyError):
            md.popitem()

    def test_setlistdefault(self):
        md = self.storage_class()
        assert md.setlistdefault("u", [-1, -2]) == [-1, -2]
        assert md.getlist("u") == [-1, -2]
        assert md["u"] == -1

    def test_iter_interfaces(self):
        mapping = [
            ("a", 1),
            ("b", 2),
            ("a", 2),
            ("d", 3),
            ("a", 1),
            ("a", 3),
            ("d", 4),
            ("c", 3),
        ]
        md = self.storage_class(mapping)
        assert list(zip(md.keys(), md.listvalues())) == list(md.lists())
        assert list(zip(md, md.listvalues())) == list(md.lists())
        assert list(zip(md.keys(), md.listvalues())) == list(md.lists())

    def test_getitem_raise_keyerror_for_empty_list_value(self):
        mapping = [("a", "b"), ("a", "c")]
        md = self.storage_class(mapping)

        md.setlistdefault("empty", [])

        with pytest.raises(KeyError):
            md["empty"]  # noqa


class TestOrderedMultiDict(_MutableMultiDictTests):
    storage_class = OrderedMultiDict  # type: ignore

    def test_ordered_interface(self):
        cls = self.storage_class

        d = cls()
        assert not d
        d.add("foo", "bar")
        assert len(d) == 1
        d.add("foo", "baz")
        assert len(d) == 2
        assert list(d.items(multi=False)) == [("foo", "bar")]
        assert list(d) == ["foo"]
        assert list(d.items()) == [("foo", "bar"), ("foo", "baz")]
        del d["foo"]
        assert not d
        assert len(d) == 0
        assert list(d) == []

        d.update([("foo", 1), ("foo", 2), ("bar", 42)])
        d.add("foo", 3)
        assert d.getlist("foo") == [1, 2, 3]
        assert d.getlist("bar") == [42]
        assert list(d.items(multi=False)) == [("foo", 1), ("bar", 42)]

        expected = ["foo", "bar"]

        assert list(d.keys()) == expected
        assert list(d) == expected
        assert list(d.keys()) == expected

        assert list(d.items()) == [
            ("foo", 1),
            ("foo", 2),
            ("bar", 42),
            ("foo", 3),
        ]
        assert len(d) == 4

        assert d.pop("foo") == 1
        assert d.pop("blafasel", None) is None
        assert d.pop("blafasel", 42) == 42
        assert len(d) == 1
        assert d.poplist("bar") == [42]
        assert not d

        assert d.get("missingkey") is None

        d.add("foo", 42)
        d.add("foo", 23)
        d.add("bar", 2)
        d.add("foo", 42)
        assert d == MultiDict(d)
        storage_id = self.storage_class(d)
        assert d == storage_id
        d.add("foo", 2)
        assert d != storage_id

        d.update({"blah": [1, 2, 3]})
        assert d["blah"] == 1
        assert d.getlist("blah") == [1, 2, 3]

        # setlist works
        d = self.storage_class()
        d["foo"] = 42
        d.setlist("foo", [1, 2])
        assert d.getlist("foo") == [1, 2]
        with pytest.raises(KeyError):
            d.pop("missing")

        with pytest.raises(KeyError):
            d["missing"]  # noqa

        # popping
        d = self.storage_class()
        d.add("foo", 23)
        d.add("foo", 42)
        d.add("foo", 1)
        assert d.popitem() == ("foo", 23)
        with pytest.raises(KeyError):
            d.popitem()
        assert not d

        d.add("foo", 23)
        d.add("foo", 42)
        d.add("foo", 1)
        assert d.popitemlist() == ("foo", [23, 42, 1])

        with pytest.raises(KeyError):
            d.popitemlist()

        # Unhashable
        d = self.storage_class()
        d.add("foo", 23)
        pytest.raises(TypeError, hash, d)

    def test_clear_clears_values(self):
        foo = self.storage_class()
        foo.add("bar", 1)
        foo.add("bar", 2)
        foo.add("foo", 3)
        foo.clear()
        assert list(foo.keys()) == []
        assert list(foo.values()) == []
        foo.add("bar", 1)
        assert dict(foo.items()) == {"bar": 1}
